#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <time.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/select.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <pthread.h>
#include <signal.h>
#include <sched.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "okc_types.h"
#include "okc_helper.h"

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif





#define DEFAULTPORT 6075 
#define DEFAULTADDRESS "192.168.10.123"
#define DEFAULTDELAY 12000000

#define SPO_QUIET 1
#define OPS_SIGNAL_PLANNER_READY 0x04

volatile sig_atomic_t keep_going = 1;
volatile sig_atomic_t late = 0;

typedef struct SERVER_PING_OPTS_T
{
  char* hostname;
  unsigned short port;
  unsigned long int delay_in_nanoseconds;
  int sockfd;
  int flags;
  pthread_t ping_thread;
  pthread_mutex_t wait_on_next_ping_mx;
  pthread_cond_t wait_on_next_ping_cd;
  timer_t *wake_ping_timer;
} server_ping_opts_t;



const char* example_package = "<Rob Type=\"KUKA\">\n"
"<RIst X=\"%f\" Y=\"%f\" Z=\"%f\" A=\"%f\" B=\"%f\" C=\"%f\"/>\n"
"<RSol X=\"%f\" Y=\"%f\" Z=\"%f\" A=\"%f\" B=\"%f\" C=\"%f\"/>\n"
"<AIPos A1=\"%f\" A2=\"%f\" A3=\"%f\" A4=\"%f\" A5=\"%f\" A6=\"%f\"/>\n"
"<ASPos A1=\"%f\" A2=\"%f\" A3=\"%f\" A4=\"%f\" A5=\"%f\" A6=\"%f\"/>\n"
"<EIPos E1=\"%f\" E2=\"0.0000\" E3=\"0.0000\" E4=\"0.0000\" E5=\"0.0000\" E6=\"0.0000\"/>\n"
"<ESPos E1=\"%f\" E2=\"0.0000\" E3=\"0.0000\" E4=\"0.0000\" E5=\"0.0000\" E6=\"0.0000\"/>\n"
"<MACur A1=\"0.0000\" A2=\"0.0000\" A3=\"0.0000\" A4=\"0.0000\" A5=\"0.0000\" A6=\"0.0000\"/>\n"
"<MECur E1=\"0.0000\" E2=\"NaN\" E3=\"NaN\" E4=\"NaN\" E5=\"NaN\" E6=\"NaN\"/>\n"
"<Torques A1=\"-42.580002\" A2=\"5.610000\" A3=\"0.270000\" A4=\"-0.090000\" A5=\"-0.160000\" A6=\"-0.030000\" E1=\"-17.250000\" X=\"0.181842\" Y=\"-0.762323\" Z=\"0.340986\" A=\"-0.046716\" B=\"-0.036019\" C=\"0.700000\" />\n"
"<Status>%d</Status>\n"
"<IPOC>%lld</IPOC>\n"
  "</Rob>\n";

static void
_print_help (const char* progname)
{
#ifdef HAVE_GETOPT_LONG
  printf ("%s %s by Matthias Sch\"opfer "
	  "%s\n\n",progname,VERSION,PACKAGE_BUGREPORT);
  printf ("USAGE: %s [OPTIONS]\n",progname);
  printf ("--help/-h\t\t\t = print this help\n");
  printf ("--quiet/-q\t\t\t = be really quiet\n");
  printf ("--port/-p <port>\t = set port address (default: %d)\n",DEFAULTPORT);
  printf ("--delay_ns/-d <delay in nanoseconds>\t\t = set delay between packets"
	  "(default: %d)\n",DEFAULTDELAY);
  printf ("--server/-s=<fqdn>\t\t = connect to server at fqdn (default: %s)\n",
	  DEFAULTADDRESS);
#else
  printf ("%s %s by Matthias Sch\"opfer "
	  "%s\n\n",progname,VERSION,PACKAGE_BUGREPORT);
  printf ("USAGE: %s [OPTIONS]\n",progname);
  printf ("-h\t\t = print this help\n");
  printf ("-q\t\t = be really quiet\n");
  printf ("-p <port> = set port address (default: %d)\n",DEFAULTPORT);
  printf ("-d <delay in nanoseconds>\t\t = set delay between packets"
	  "(default: %d)\n",DEFAULTDELAY);
  printf ("-s=<fqdn>\t\t = connect to server at fqdn (default: %s)\n",
	  DEFAULTADDRESS);
#endif
}

void
catch_interrupt (int sig)
{
  keep_going = 0;
  signal (sig, catch_interrupt);
  
}


/* get sockaddr, IPv4 or IPv6:*/
static void 
*get_in_addr(struct sockaddr *sa)
{
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

static void
fill_spo_defaults (server_ping_opts_t* spo)
{
  spo->hostname = malloc (strlen (DEFAULTADDRESS)+1);
  
  if (NULL == spo->hostname)
    {
      perror ("malloc");
      exit (EXIT_FAILURE);
    }

  strcpy (spo->hostname,DEFAULTADDRESS);

  spo->port = DEFAULTPORT;
  spo->delay_in_nanoseconds = DEFAULTDELAY;
  spo->flags = 0;
}

static void
init_server_ping_opts (server_ping_opts_t* spo)
{
  spo->wake_ping_timer = malloc (sizeof (timer_t));
  if (spo->wake_ping_timer == NULL)
    {
      perror ("okc_ping_server: init_server_ping_opts: malloc");
      exit (EXIT_FAILURE);
    }

  if (pthread_mutex_init (&spo->wait_on_next_ping_mx, NULL) != 0)
    {
      perror ("okc_ping_server: init_server_ping_opts: pthread_mutex_init");
      exit (EXIT_FAILURE);
    }

  if (pthread_cond_init (&spo->wait_on_next_ping_cd, NULL) != 0)
    {
      perror ("okc_ping_server: init_server_ping_opts: pthread_cond_init");
      exit (EXIT_FAILURE);
    }
  spo->sockfd = -1;
}
  
static server_ping_opts_t*
parse_args (int argc, char** argv)
{
#ifdef HAVE_GETOPT_LONG
  static struct option long_options[] = {
    {"help", no_argument, 0, 'h'},
    {"quiet", no_argument, 0, 'q'},
    {"port", required_argument, 0, 'p'},
    {"delay", required_argument, 0, 'd'},
    {"server",required_argument,0,'s'},
    {0, 0, 0, 0}
  };
#endif

  char* optstring = "hqp:d:s:";
  char* progname;
  
  int ret; 
  
#ifdef HAVE_GETOPT_LONG
  int longindex;
#endif
  server_ping_opts_t* spo;

  spo = malloc (sizeof (server_ping_opts_t));

  if (NULL == spo)
    {
      perror ("malloc");
      exit (EXIT_FAILURE);
    }

  init_server_ping_opts (spo);

  progname = malloc (strlen (argv[0])+1);
  if (NULL == progname)
    {
      perror ("malloc");
      exit (EXIT_FAILURE);
    }

  strcpy (progname,argv[0]);
  

  fill_spo_defaults (spo);

  #ifdef  HAVE_GETOPT_LONG 
  while ((ret = getopt_long_only (argc, argv, optstring, long_options, 
				  &longindex)) != -1)
#else
    while ((ret = getopt (argc,argv,optstring)) != -1)
#endif
    {
      if (ret == '?' || ret == 'h')
	{
	  _print_help (progname);
	  exit (EXIT_FAILURE);
	}
      switch (ret)
	{
	case 'q':
	  spo->flags |= SPO_QUIET;
	  break;
	case 'p':
	  if (optarg != NULL)
	    spo->port = atoi(optarg);
	  else
	    {
	      printf ("Option port requires an argument!\n");
	      _print_help (progname);
	      exit (EXIT_FAILURE);
	    }
	  break;
	case 'd':
	  if (optarg != NULL)
	    spo->delay_in_nanoseconds = atoi (optarg);
	   else
	    {
	      printf ("Option delay requires an argument!\n");
	      _print_help (progname);
	      exit (EXIT_FAILURE);
	    }
	  break;
	case 's':	 
	  if (optarg != NULL)
	    {
	      free (spo->hostname);
	      spo->hostname = malloc (strlen (optarg)+1);
	      if (NULL == spo->hostname)
		{
		  perror ("malloc");
		  exit (EXIT_FAILURE);
		}
	      strcpy (spo->hostname,optarg);
	    }
	  else
	    {
	      printf ("Option server requires an argument!\n");
	      _print_help (progname);
	      exit (EXIT_FAILURE);
	    }
	  break;
	default:
	  _print_help (progname);
	  exit (EXIT_FAILURE);
	}
    }
  
  free(progname);
  return spo;
}

int
connect_to_server (server_ping_opts_t* spo)
{
  int sockfd;
  struct addrinfo hints, *servinfo, *p;
  char s[INET6_ADDRSTRLEN];
  int rv;

  char portnumber[64];

  snprintf (portnumber,64,"%u",spo->port);
  
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  
  if ((rv = getaddrinfo(spo->hostname,portnumber, &hints, &servinfo)) != 0) 
    {
      fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
      exit (EXIT_FAILURE);
    }
  
  // loop through all the results and connect to the first we can
  for(p = servinfo; p != NULL; p = p->ai_next) 
    {
      if ((sockfd = socket(p->ai_family, p->ai_socktype,
			   p->ai_protocol)) == -1) 
	{
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	    perror("socket");
	  continue;
	}
      
      if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1) 
	{
	  close(sockfd);
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	    perror("connect");
	  continue;
	}
      
      break;
    }
  
  if (p == NULL) 
    {
      fprintf(stderr, "failed to connect\n");
      exit (EXIT_FAILURE);
    }
  
  inet_ntop(p->ai_family, get_in_addr((struct sockaddr *)p->ai_addr),
            s, sizeof s);
  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
    printf("connecting to %s\n", s);
  
  freeaddrinfo(servinfo);
  return sockfd;
}



/**
 * Helper function intern to OpenKC. Parses XML-Section related to long-int-set.
 *
 * @param node #xmlNode that holds long int data.
 *
 * @param llip pointer to long long int to be filled.
 *
 * @return EXIT_SUCCESS on success, EXIT_FAILURE on failure.
 */
static int
_parse_text_to_long_long_int_node (xmlNode* node, long long int* llip)
{
  unsigned char* buffer = NULL;
  unsigned char* detect_error = NULL;

  int outlen, inlen;
  int ret;

  outlen = inlen = xmlUTF8Strlen (node->children->content);

  buffer = malloc ((inlen+1)*sizeof (char));

  buffer[inlen] = '\000';

  ret = UTF8Toisolat1 (buffer,&outlen,node->children->content, 
		 &inlen);
  
  (*llip) = strtoll ((char*) buffer, (char**) &detect_error, 10);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in IPOC\n",buffer);
      return (EXIT_FAILURE);
    }
  free (buffer);
  return (EXIT_SUCCESS);
}

static int
_parse_axis_node (xmlNode* node, lbr_axis* gen_axis)
{
  char* buffer;
  char* detect_error;
  
  if ((NULL == (buffer = (char*) xmlGetProp (node, BAD_CAST "A1"))) && (NULL == (detect_error = (char*) xmlGetProp (node, BAD_CAST "E1"))))
    {
      fprintf (stderr,"parse_tree(axis): wrong node?!\n");
      return (EXIT_FAILURE); 
    }

  if (NULL != buffer)
    {
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A1");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A1 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a1 = strtod (buffer,&detect_error);
      
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A1\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A2");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A2 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}

      gen_axis->a2 = strtod (buffer,&detect_error);
  
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A2\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A3");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A3 where it "
		   "should be!\n");

	  return (EXIT_FAILURE);
	}

      gen_axis->a3 = strtod (buffer,&detect_error);
  
      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A3\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A4");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A4 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a4 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A4\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A5");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A5 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}
      
      gen_axis->a5 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A5\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "A6");

      if (NULL == buffer)
	{
	  fprintf (stderr,"parse_tree: Did not find Property A6 where it "
		   "should be!\n");
	  return (EXIT_FAILURE);
	}

      gen_axis->a6 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property A6\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
    }
  else
    {
      xmlFree (detect_error);
      detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "E1");

      if (NULL == buffer)
	{
	  fprintf (stderr,"Did not find Property E1 where it should be!\n");
	  return (EXIT_FAILURE);
	}
		 
      gen_axis->e1 = strtod (buffer,&detect_error);

      if (buffer == detect_error)
	{
	  fprintf (stderr,"parse_tree: String conversion failed while trying to "
		   "parse '%s' in gen_axis, Property E1\n",buffer);
	  xmlFree (buffer);
	  return (EXIT_FAILURE);
	}
      xmlFree (buffer);
    }
  return (EXIT_SUCCESS);
}

static int
_parse_param_node (xmlNode* node, lbr_axis* gen_axis)
{
  char* buffer;
  char* detect_error;
  
  if (NULL == (buffer = (char*) xmlGetProp (node, BAD_CAST "P1")))
    {
      fprintf (stderr,"parse_tree (param): wrong node?!\n");
      return (EXIT_FAILURE); 
    }

  
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P1");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P1 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
      
  gen_axis->a1 = strtod (buffer,&detect_error);
      
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P1\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P2");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P2 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_axis->a2 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P2\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P3");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P3 where it "
	       "should be!\n");

      return (EXIT_FAILURE);
    }

  gen_axis->a3 = strtod (buffer,&detect_error);
  
  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P3\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P4");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P4 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
      
  gen_axis->a4 = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P4\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P5");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P5 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }
      
  gen_axis->a5 = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P5\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P6");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P6 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_axis->a6 = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P6\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  detect_error = buffer = (char*) xmlGetProp (node, BAD_CAST "P7");

  if (NULL == buffer)
    {
      fprintf (stderr,"parse_tree: Did not find Property P7 where it "
	       "should be!\n");
      return (EXIT_FAILURE);
    }

  gen_axis->e1 = strtod (buffer,&detect_error);

  if (buffer == detect_error)
    {
      fprintf (stderr,"parse_tree: String conversion failed while trying to "
	       "parse '%s' in gen_axis, Property P7\n",buffer);
      xmlFree (buffer);
      return (EXIT_FAILURE);
    }
  xmlFree (buffer);
  return (EXIT_SUCCESS);
}


static int
_parse_tree (xmlDoc* doc, long long int* ipoc, lbr_axis* gen_axis, 
	     long long int* command, lbr_axis* params)
{
  xmlNode* root_node = NULL;
  xmlNode* cur_node = NULL;
  
  
  root_node = xmlDocGetRootElement(doc);
  
  if (NULL == root_node)
    {
      fprintf (stderr,"_parse_tree: Error parsing doctree!\n");
      return (EXIT_FAILURE);
    }

  if (0 != xmlStrcmp (root_node->name,BAD_CAST "Sen"))
    return (EXIT_FAILURE);

  cur_node = root_node->xmlChildrenNode;
  
  while (NULL != cur_node)
    {
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "IPOC")))
	{
	  if (EXIT_SUCCESS != _parse_text_to_long_long_int_node (cur_node, ipoc))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "AKorr")))
	{
	   if (EXIT_SUCCESS != _parse_axis_node (cur_node, gen_axis))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "EKorr")))
	{
	   if (EXIT_SUCCESS != _parse_axis_node (cur_node, gen_axis))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "Command")))
	{
	   if (EXIT_SUCCESS != _parse_text_to_long_long_int_node (cur_node, command))
	    return (EXIT_FAILURE);
	}
      if ((cur_node->type == XML_ELEMENT_NODE) && 
	  (0 == xmlStrcmp (cur_node->name,BAD_CAST "Param")))
	{
	   if (EXIT_SUCCESS != _parse_param_node (cur_node, params))
	    return (EXIT_FAILURE);
	}
      cur_node = cur_node->next;
    }
  return (EXIT_SUCCESS);
}

static size_t
calc_time (struct timespec start, struct timespec stop)
{
  size_t ret;
  if (start.tv_sec != stop.tv_sec)
    {
      ret = (1000000000 - start.tv_nsec) + stop.tv_nsec;
    }
  else
    {
      ret = stop.tv_nsec - start.tv_nsec;
    }

  return ret;
}

static void
_ops_timer_intr (int sig, siginfo_t * extra, void *cruft)
{
  int noverflow;
  server_ping_opts_t *spo = extra->si_value.sival_ptr;
  int errorno;


  if ((noverflow = timer_getoverrun (*(spo->wake_ping_timer))))
    {
      late++;
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr, "okc_server_ping: warning: timer overflow by %d\n", 
		 noverflow);
    }
  else
    {
      if ((errorno = pthread_mutex_trylock (&spo->wait_on_next_ping_mx)) != 0)
	{
	  if (errorno == EBUSY)
	    {
	      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
		fprintf (stderr, "okc_ping_server: warning: ping task not ready,"
			 " skipping (trylock)\n");
	      late++;
	    }
	  else
	    {
	      perror ("okc_ping_server: pthread_mutex_trylock");
	      exit (EXIT_FAILURE); 
	    }
	}
      else
	{
	  if ((spo->flags & OPS_SIGNAL_PLANNER_READY) == OPS_SIGNAL_PLANNER_READY)
	    {
	      if (0 != 
		  pthread_cond_signal (&spo->wait_on_next_ping_cd))
		{
		  perror ("okc_ping_server: _ops_timer_intr: "
			  "pthread_cond_signal");
		  exit (EXIT_FAILURE);
		}
	      if (0 != 
		  pthread_mutex_unlock (&spo->wait_on_next_ping_mx ))
		{
		  perror ("okc_ping_server: _ops_timer_intr: "
			  "pthread_mutex_unlock");
		  exit (EXIT_FAILURE);
		}
	    }
	  else
	    {
	      late++;
	      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
		
		fprintf (stderr, "okc_ping_server: warning: ping task not "
			 "ready, skipping (flag)\n");
	      if (0 != 
		  pthread_mutex_unlock (&spo->wait_on_next_ping_mx ))
		{
		  perror ("okc_ping_server: _ops_timer_intr: "
			  "pthread_mutex_unlock");
		  exit (EXIT_FAILURE);
		}
	    }
	}
    }
  return;
}

static float
time_passed_in_secs (struct timeval start, struct timeval stop)
{
  float ms=0.0;


  ms = (float) (stop.tv_sec - start.tv_sec);  

  if (stop.tv_usec > start.tv_usec)
    ms += ((float) stop.tv_usec - start.tv_usec) / 1000000.0;
  else
    ms += ((float)(1000000 - start.tv_usec) + (stop.tv_usec))/1000000.0 - 1.0;
  return ms;
}

  

static void*
ping_server (void* spo_ptr)
{    
  server_ping_opts_t* spo = spo_ptr;
  struct timeval wait;
  struct timeval start_all;
  struct timeval stop_all;
  struct timespec start_time,stop_time;
  struct timeval start_gtod, stop_gtod,last_gtod;
  ssize_t nbytes;
  fd_set readfds;
  long long int ipoc,ret_ipoc,ipoc_start;
  char* buffer;
  size_t size;
  xmlDoc* doc = NULL;
  size_t good, bad;
  long long int nsecs_total;
  size_t min_nsecs, max_nsecs;
  size_t nanosecs_passed;
  float secs_total = 0.0;
  float secs_passed;
  float min_secs,max_secs;
  sigset_t allsigs;
  double joints[7] = {78.8004,127.1174,-61.2806,86.4107,105.2399,-144.9521, 0.0};
  double position[6] = {33.1614,-603.7832,567.8233,92.1401,-44.1952,92.3466};
  lbr_axis akorr;
  lbr_axis param;
  long long int command;
  int status=0;
  

  sigfillset (&allsigs);
  pthread_sigmask (SIG_BLOCK, &allsigs, NULL);

  gettimeofday (&start_all,NULL);
  gettimeofday (&stop_gtod,NULL);
  max_nsecs = nsecs_total = good = bad = 0;
  min_nsecs = spo->delay_in_nanoseconds;
  max_secs = 0.0;
  min_secs = 1.0;

  ret_ipoc = 0;
  ipoc_start = ipoc  = start_all.tv_sec;
  last_gtod = start_all;
  size = strlen (example_package) + 300;
  buffer = malloc (size);
  if (NULL == buffer)
    {
      perror ("malloc");
      exit (EXIT_FAILURE);
    }

  while (keep_going)
    {
      ipoc++;
      snprintf (buffer,size,example_package,position[0],position[1],
		position[2],position[3],position[4],position[5],
		position[0],position[1],position[2],position[3],
		position[4],position[5],joints[0],joints[1],joints[2],
		joints[3],joints[4],joints[5],joints[0],joints[1],joints[2],
		joints[3],joints[4],joints[5],joints[6],joints[6],status,
		ipoc);
      
      if (0 != pthread_mutex_lock (&spo->wait_on_next_ping_mx))
	{
	  perror ("okc_ping_server: ping_server: pthread_mutex_lock");
	  exit (EXIT_FAILURE);
	}
      /* Signal Interrupt Handler that we are ready if he is... */
      spo->flags |= OPS_SIGNAL_PLANNER_READY;

      /* And wait for Interrupt Handler to signal us. */
      if (0 != pthread_cond_wait (&spo->wait_on_next_ping_cd,
				  &spo->wait_on_next_ping_mx)) 
	{
	  perror ("okc_ping_server: ping_server: pthread_cond_wait");
	  exit (EXIT_FAILURE);
	}
      spo->flags ^= OPS_SIGNAL_PLANNER_READY;
      /* Release lock */
      if (0 != pthread_mutex_unlock (&spo->wait_on_next_ping_mx))
	{
	  perror ("okc_ping_server: ping_server: pthread_mutex_unlock");
	  exit (EXIT_FAILURE);
	}
      gettimeofday (&start_gtod,NULL);
      clock_gettime (CLOCK_REALTIME,&start_time);
      send (spo->sockfd,buffer,strlen(buffer),0);
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr,"%lu bytes send after %f seconds passed (gtod) ",strlen(buffer),
		 time_passed_in_secs (last_gtod,start_gtod));
      last_gtod = start_gtod;
      FD_ZERO (&readfds);
      FD_SET (spo->sockfd,&readfds);
      wait.tv_sec = 0;
      wait.tv_usec = spo->delay_in_nanoseconds;
      if (-1 == select (spo->sockfd+1,&readfds,NULL,NULL,&wait))
	{
	  perror ("select");
	  exit (EXIT_FAILURE);
	}
      nbytes = recv (spo->sockfd,buffer,size,MSG_DONTWAIT);
      clock_gettime (CLOCK_REALTIME,&stop_time);
      gettimeofday (&stop_gtod,NULL);
      
      nanosecs_passed = calc_time (start_time,stop_time);
      secs_passed = time_passed_in_secs (start_gtod,stop_gtod);
      if (nanosecs_passed > max_nsecs)
	max_nsecs = nanosecs_passed;
      if (nanosecs_passed < min_nsecs)
	min_nsecs = nanosecs_passed;

      if (secs_passed > max_secs)
	max_secs = secs_passed;
      if (secs_passed < min_secs)
	min_secs = secs_passed;

      if (nbytes == 0)
	{
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	    fprintf (stderr," - timeout\n");
	}
      else
	{
	  doc = xmlReadMemory (buffer, nbytes ,"noname.xml",NULL,0);
	  if (EXIT_SUCCESS != _parse_tree (doc, &ret_ipoc, &akorr,
					   &command, &param))
	    {
	      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
		fprintf (stderr,"error parsing answer: \n%s\n",
			 buffer);
	    }
	  else
	    {
	      joints[0] += akorr.a1;
	      joints[1] += akorr.a2;
	      joints[2] += akorr.a3;
	      joints[3] += akorr.a4;
	      joints[4] += akorr.a5;
	      joints[5] += akorr.a6;
	      joints[6] += akorr.e1;

	      switch (command)
		{
		case OKC_SET_CP_STIFFNESS: 
		  {
		    printf ("Received OKC_SET_CP_STIFFNESS. Params:\n");
		    okc_print_axis (param);
		    status = OKC_ACK_SET_CP_STIFFNESS;
		    break;
		  }
		case OKC_SET_CP_DAMPING:
		  {
		    printf ("Received OKC_SET_CP_DAMPING. Params:\n");
		    okc_print_axis (param);
		    status = OKC_ACK_SET_CP_DAMPING;
		    break;
		  }
		case OKC_SET_AXIS_STIFFNESS:
		  {
		    printf ("Received OKC_SET_AXIS_STIFFNESS. Params:\n");
		    okc_print_axis (param);
		    status = OKC_ACK_SET_AXIS_STIFFNESS;
		    break;
		  }
		case OKC_SET_AXIS_DAMPING:
		  {
		    printf ("Received OKC_SET_AXIS_DAMPING. Params:\n");
		    okc_print_axis (param);
		    status = OKC_ACK_SET_AXIS_DAMPING;
		    break;
		  }
		case OKC_SET_SWITCH_CP_IMPEDANCE:
		  {
		    printf ("Received OKC_SET_SWITCH_CP_IMPEDANCE.\n");
		    status = OKC_ACK_SET_SWITCH_CP_IMPEDANCE;
		    break;
		  }
		case OKC_SET_SWITCH_AXIS_IMPEDANCE:
		  {
		    printf ("Received OKC_SET_SWITCH_AXIS_IMPEDANCE.\n");
		    status = OKC_ACK_SET_SWITCH_AXIS_IMPEDANCE;
		    break;
		  }
		case OKC_SET_SWITCH_GRAVCOMP:
		  {
		    printf ("Received OKC_SET_SWITCH_GRAVCOMP.\n");
		    status = OKC_ACK_SET_SWITCH_GRAVCOMP;
		    break;
		  }
		case OKC_SET_SWITCH_POSITION:
		  {
		    printf ("Received OKC_SET_SWITCH_POSITION.\n");
		    status = OKC_ACK_SET_SWITCH_POSITION;
		    break;
		  }
		case OKC_SET_EXIT_GRACEFULLY:
		  {
		    printf ("Received OKC_SET_SWITCH_POSITION.\n");
		    keep_going = 0;
		    break;
		  }
		case OKC_NORMAL_OP:
		  {
		    status = OKC_OK;
		    break;
		  }
		case OKC_COMMAND_PENDING:
		  {
		    printf ("Received OKC_COMMAND_PENDING.\n");
		    break;
		  }
		default:
		  {
		    printf ("Received unknown Command: %lld\n",command);
		  }
		}
	    }
	  xmlFreeDoc (doc);
	  if ((ret_ipoc) == ipoc && (late == 0))
	    {
	      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
		fprintf (stderr," - received answer after %lu ns (RT) or %f secs (GTOD), ipoc correct\n",
			 nanosecs_passed, time_passed_in_secs (start_gtod,stop_gtod));
	      nsecs_total += nanosecs_passed;
	      secs_total += secs_passed;
	      good++;
	    }
	  else
	    {
	      if (late)
		bad += late;
	      else
		bad++;
	      late = 0;
	      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
		fprintf (stderr,
			 " - received answer after %lu ns (RT) or %f secs (GTOD), ipoc NOT correct (%lld vs %lld)\n",
			 nanosecs_passed,time_passed_in_secs (start_gtod,stop_gtod),ret_ipoc,ipoc);
	    }
	  
	}
     
    }
  gettimeofday (&stop_all, NULL);

  clock_getres (CLOCK_REALTIME,&start_time);

  printf ("\nStatistics: Packets Send Total %lld\n"
	  "            Packets Received in time %lu (%f %%)\n"
	  "            Late/Lost Packets %lu (%f %%)\n"
	  "            Average time to answer (CLOCK_REALTIME) %f ms (max: %f ms, min %f ms)\n"
	  "            Average time to answer (gettimeofday) %f secs (max: %f secs, min %f secs)\n"
	  "            Ping server had run %f seconds, which results in %f frames per second\n"
	  "            CLOCK_REALTIME says its Timer Resolution is %lu nanosecs\n",
	  ipoc - ipoc_start, good, (float)good*100.0 / (ipoc - ipoc_start),
	  bad, (float) bad*100.0 / (ipoc - ipoc_start),
	  ((float) nsecs_total / (float)good) / 1000000.0,
	  (float) max_nsecs / 1000000.0, 
	  (float) min_nsecs / 1000000.0,
	  secs_total / (float) good,
	  max_secs,
	  min_secs,
	  time_passed_in_secs (start_all, stop_all),
	  (float) (ipoc - ipoc_start) / time_passed_in_secs (start_all, stop_all),
	  start_time.tv_nsec);
  free (buffer);
  return NULL;
}

void
start_ping_thread (server_ping_opts_t* spo)
{
  struct timespec tp;
  struct sigaction sa;
  struct itimerspec i;
  struct sigevent timer_event;
  pthread_attr_t thread_attr;
  struct sched_param param;
  int result, policy;
  

  sa.sa_flags = SA_SIGINFO;	/*real-time signal */
  sa.sa_sigaction = _ops_timer_intr;	/*pointer to action */

  if (sigaction (SIGRTMIN, &sa, NULL) < 0)
    {
      perror ("okc_ping_server: start_ping_thread: sigaction error");
      exit (EXIT_FAILURE);
    }

  /* catch CTRL-C and such */
  signal (SIGINT, catch_interrupt);
  signal (SIGTERM, catch_interrupt);


  /*
   * first determine whether the desired clock exists
   */

  if (clock_getres (CLOCK_REALTIME, &tp) != 0)
    {
      perror ("okc_ping_server: start_ping_thread: clock_getres");
      exit (EXIT_FAILURE);
    }

  if (2 * tp.tv_nsec > spo->delay_in_nanoseconds)
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	{
	  fprintf (stderr,
		   "okc_ping_server: start_ping_thread:: warning: clock "
		   "resolution of system probably "
		   "\n\t\tinsufficient.\n");
	  fprintf (stderr, "\t\tclock resolution:\t %ld sec, %ld nanosec.\n",
		   tp.tv_sec, tp.tv_nsec);
	  fprintf (stderr, "\t\tdesired intervall:\t %ld sec, %ld nanosec.\n",
		   (long int) 0, spo->delay_in_nanoseconds);
	  
	  fprintf (stderr, "\t\ttrying to continue anyway... good luck!\n");
	}
    }

  /*
   * create a timer based upon the CLOCK_REALTIME clock 
   */

  /* set resolution to the desired values */

  i.it_interval.tv_nsec = spo->delay_in_nanoseconds;
  i.it_interval.tv_sec = 0;
  i.it_value = i.it_interval;

  /*
   * this describes the asynchronous notification to be posted
   * upon this timer's expiration:
   *
   * - use signals
   * - send SIGRTMIN
   * - send extra data consisting of a pointer back to the timer ID
   *   cannot pass the timer ID itself because we haven't created it
   *   yet.
   */

  timer_event.sigev_notify = SIGEV_SIGNAL;
  timer_event.sigev_signo = SIGRTMIN;
  timer_event.sigev_value.sival_ptr = (void *) spo;

  if (timer_create (CLOCK_REALTIME, &timer_event, spo->wake_ping_timer) <
      0)
    {
      perror ("okc_ping_server: start_ping_thread: timer_create");
      exit (EXIT_FAILURE);
    }

  /* relative timer, go off at the end of the interval */

  if (timer_settime (*(spo->wake_ping_timer), 0, &i, NULL) < 0)
    {
      perror ("okc_ping_server: start_ping_thread: timer_settime");
      exit (EXIT_FAILURE);
    }
  
  if (0 != pthread_attr_init (&thread_attr))
    {
      perror ("okc_ping_server: pthread_attr_init");
      exit (EXIT_FAILURE);
    }

  if (0 != pthread_attr_setschedpolicy (&thread_attr, SCHED_FIFO))
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr,"okc_ping_server: warning: no SCHED_FIFO available\n");
    }
  
  

  param.sched_priority = 50;
  if (0 != pthread_attr_setschedparam(&thread_attr, &param))
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr,"okc_ping_server: warning: no SCHED_PRIORITY "
		 "available\n");
    }

  if (0 != pthread_create (&spo->ping_thread, &thread_attr, 
			   ping_server, (void*) spo))
    {
      perror ("okc_ping_server: pthread_create");
      exit (EXIT_FAILURE);
    }

  if (0 != (result = pthread_getschedparam (spo->ping_thread,&policy,
					    &param)))
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	{
	  perror ("okc_ping_server: pthread_getschedparam");
	  fprintf (stderr,"result was %d\n",result);
	}
    }
  else
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr,"okc_ping_server: info: thread is in class "
		 "%s at priority %d\n",
		 (policy==SCHED_OTHER)?"SCHED_OTHER":(policy==SCHED_FIFO)?
		 "SCHED_FIFO":"SCHED_RR", param.sched_priority);
    }

  if (policy != SCHED_FIFO)
    {
      if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	fprintf (stderr,"okc_ping_server: info: altering thread priority\n");
      policy = SCHED_FIFO;
      param.sched_priority = 50;
      if (0 != pthread_setschedparam (spo->ping_thread,policy,&param))
	{
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	    {
	      perror ("okc_ping_server: pthread_setschedparam");
	      fprintf (stderr,"okc_ping_server: warning: failed to set up "
		       "realtime task\n");
	    }
	}
      
      if (0 != (result = pthread_getschedparam (spo->ping_thread,&policy,
						&param)))
	{
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)
	    {
	      perror ("okc_ping_server: pthread_getschedparam");
	      fprintf (stderr,"result was %d\n",result);
	    }
	}
      else
	{
	  if ((spo->flags & SPO_QUIET) != SPO_QUIET)

	    fprintf (stderr,"okc_ping_server: info: thread is in class %s "
		     "at priority %d\n",
		     (policy==SCHED_OTHER)?"SCHED_OTHER":(policy==SCHED_FIFO)?
		     "SCHED_FIFO":"SCHED_RR", param.sched_priority);
	}
    }

  pthread_attr_destroy (&thread_attr);
}

static void
pimp_me_to_rt (void)
{
  pid_t own_pid;
  struct sched_param param;
  int error;

  own_pid = getpid ();



  param.sched_priority = 55;
  
  if (0 != (error = sched_setscheduler (own_pid,SCHED_FIFO,&param)))
    {
      fprintf (stderr,"failed to alter scheduling parameters\n");
      perror ("sched_setparam");
    }
}
	
  

int 
main(int argc, char **argv)
{
  server_ping_opts_t* spo;

  LIBXML_TEST_VERSION;

  pimp_me_to_rt();

  xmlInitParser();

  spo = parse_args (argc,argv);
  
  spo->sockfd = connect_to_server (spo);
  
  start_ping_thread (spo);
  
  pthread_join (spo->ping_thread,NULL);

  //atexit (print_stats);
  close(spo->sockfd);

  free (spo->hostname);
  free (spo);

  return 0;
}
